const preferNetworkList = ["/"];
const WEB_PLAYER_ASSETS_CACHE = "pixelart-player-files-0.192.0";
const isDevelopment = WEB_PLAYER_ASSETS_CACHE.endsWith("VERSION");

const certificateHashesUrl =
  "https://apps-api.fugo.ai/on-premise-certificates/client";
const CERTIFICATE_EXPIRATION_BUFFER = 1000 * 60 * 60 * 24; // 1 day; API generates new certificates 2 days before expiration

// eslint-disable-next-line no-restricted-globals
self.addEventListener("install", function (event) {
  const preCache = async () => {
    try {
      console.log("fetching manifest");
      const resp = await fetch("/asset-manifest.json");
      const reply = await resp.json();
      const files = Object.entries(reply.files)
        .filter(([key]) => !key.endsWith("map"))
        .filter(([key]) => key !== "index.html")
        .map(([_, url]) => url);
      console.log({ files });
      const cache = await caches.open(WEB_PLAYER_ASSETS_CACHE);
      await cache.addAll(["/", ...files]);
    } catch (e) {
      console.error("SW install error", e);
    }
  };

  // eslint-disable-next-line no-restricted-globals
  if (!isDevelopment) event.waitUntil(preCache());

  // eslint-disable-next-line no-restricted-globals
  self.skipWaiting();
});

// eslint-disable-next-line no-restricted-globals
self.addEventListener("activate", function (event) {
  // eslint-disable-next-line no-restricted-globals
  event.waitUntil(self.clients.claim());

  const cacheKeeplist = [WEB_PLAYER_ASSETS_CACHE, "pixelart-files"];

  event.waitUntil(
    caches.keys().then((keyList) => {
      return Promise.all(
        keyList.map((key) => {
          if (cacheKeeplist.indexOf(key) === -1) {
            return caches.delete(key);
          }
        })
      );
    })
  );
});

// eslint-disable-next-line no-restricted-globals
self.addEventListener("fetch", (event) =>
  event.respondWith(getResponse(event))
);

let webTransportProxyHostname = "";
let webTransportProxyHost = "";
async function getResponse(event) {
  if (event.request.url.endsWith("/web-transfer-proxy")) {
    const { host, hostname } = new URL(event.request.url);
    webTransportProxyHostname = hostname;
    webTransportProxyHost = host;
    const url = `https://${host}/wt`;
    let client = await createWebTransportClient(url);
    try {
      return await retryWithTimeout(
        async () => await proxyGraphQLRequest(event, client),
        1,
        10000,
        async () => {
          try {
            client.close();
          } catch (e) {
            console.error(e);
          }
          client = await createWebTransportClient(url);
        }
      );
    } catch (e) {
      client.close();
      throw e;
    }
  } else if (
    webTransportProxyHostname != null &&
    event.request.url.includes(webTransportProxyHostname) &&
    event.request.url.includes("/screenshots/")
  ) {
    const url = `https://${webTransportProxyHost}/wt`;
    console.log({ url });
    let client = await createWebTransportClient(url);
    try {
      return await retryWithTimeout(
        async () => await proxyImageRequest(event, client),
        1,
        10000,
        async () => {
          try {
            client.close();
          } catch (e) {
            console.error(e);
          }
          client = await createWebTransportClient(url);
        }
      );
    } catch (e) {
      client.close();
      throw e;
    }
  } else if (getIsPreferNetwork(event)) {
    try {
      return await fetchAndCache(event);
    } catch (e) {
      console.error(`Failed to fetch ${event.request.url}`, e);
    }
  }

  const cached = await getFromCache(event);
  if (cached) return cached;

  return await fetchAndCache(event);
}

async function proxyGraphQLRequest(event, client) {
  const start = Date.now();
  const stream = await client.createBidirectionalStream();
  const readable = stream.readable;
  const message = await event.request.clone().text();
  await sendWebTransportMessage(message, stream);
  const reader = readable.getReader();
  let reply = "";
  const decoder = new TextDecoder("utf-8");
  let start1 = Date.now();
  while (true) {
    const { value, done } = await reader.read();
    if (done) {
      break;
    }
    const chunk = decoder.decode(value).toString();
    reply += chunk;
  }
  console.log(`%c read graphql in ${Date.now() - start1}ms, ${reply}`);
  const response = new Response(reply, {
    status: 200,
    statusText: "OK",
    ok: true,
    headers: {
      "Content-Type": "application/json",
    },
  });
  client.close();
  console.log(`proxyGraphQLRequest took ${Date.now() - start}ms`);
  return response;
}

async function proxyImageRequest(event, client) {
  const start = Date.now();
  const stream = await client.createBidirectionalStream();
  const readable = stream.readable;
  await sendWebTransportMessage(event.request.url, stream);
  const reader = readable.getReader();
  let chunks = [];
  const start1 = Date.now();
  while (true) {
    const { value, done } = await reader.read();
    if (done) {
      break;
    }
    chunks.push(value);
  }
  console.log(`%c read image in ${Date.now() - start1}ms`);
  const reply = new Blob(chunks);
  const response = new Response(reply, {
    status: 200,
    statusText: "OK",
    ok: true,
    headers: {
      "Content-Type": "image/png",
    },
  });
  client.close();
  console.log(`proxyImageRequest took ${Date.now() - start}ms`);
  return response;
}

async function sendWebTransportMessage(message, stream) {
  const msg = `${message.length} ${message}`;
  const writer = stream.writable.getWriter();
  const encoder = new TextEncoder();
  const encoded = encoder.encode(msg);
  const start = Date.now();
  console.log("waiting for the writer to be ready");
  await writer.ready;
  console.log("writer is ready");
  await writer.write(encoded);
  console.log(`wrote. waiting for the writer to be ready ${msg}`);
  await writer.ready;
  console.log("writer is ready. about to close it");
  await writer.close();
  console.log(`writer is closed. Wrote in ${Date.now() - start}ms`);
}

async function retryWithTimeout(promise, tries, ms, cleanUpFn) {
  for (let i = 0; i < tries; ++i) {
    try {
      return await timeout(promise, ms * 2);
    } catch (e) {
      console.log(e, i);
      await cleanUpFn();
    }
    const rnd = Math.floor(Math.random() * (400 - 50 + 1) + 50);
    console.log(`sleep for ${rnd}ms`);
    await new Promise((resolve) => setTimeout(resolve, rnd));
  }
  throw new Error(`Retried ${tries} times with ${ms} timeout and failed.`);
}

function timeout(promise, ms) {
  return new Promise((resolve, reject) => {
    let hasTimeouted = false;
    const timer = setTimeout(() => {
      hasTimeouted = true;
      reject(new Error("TIMEOUT"));
    }, ms);
    promise()
      .then((value) => {
        clearTimeout(timer);
        if (!hasTimeouted) {
          resolve(value);
        }
      })
      .catch((reason) => {
        clearTimeout(timer);
        if (!hasTimeouted) {
          reject(reason);
        }
      });
  });
}

const STORE_NAME = "webTransportCertificates";

let db;

const openDB = () => {
  return new Promise((resolve, reject) => {
    const request = indexedDB.open("webTransportCache", 1);

    request.onerror = function (event) {
      reject("Error opening db", event.target.errorCode);
    };

    request.onsuccess = function (event) {
      resolve(event.target.result);
    };

    request.onupgradeneeded = function (event) {
      const db = event.target.result;
      if (!db.objectStoreNames.contains(STORE_NAME)) {
        db.createObjectStore(STORE_NAME);
      }
    };
  });
};

const getWebTransportCacheDB = async () => {
  if (db) {
    return db;
  }
  db = await openDB();
  return db;
};

async function setValueToIndexedDB(key, value) {
  const db = await getWebTransportCacheDB();

  return new Promise((resolve, reject) => {
    const transaction = db.transaction([STORE_NAME], "readwrite");
    const store = transaction.objectStore(STORE_NAME);
    const request = store.put(value, key);

    request.onerror = function (event) {
      reject("Error writing value", event.target.errorCode);
    };

    request.onsuccess = function (event) {
      resolve();
    };
  });
}

async function getValueFromIndexedDB(key) {
  const db = await getWebTransportCacheDB();

  return new Promise((resolve, reject) => {
    const transaction = db.transaction([STORE_NAME], "readonly");
    const store = transaction.objectStore(STORE_NAME);
    const request = store.get(key);

    request.onerror = function (event) {
      reject("Error fetching value", event.target.errorCode);
    };

    request.onsuccess = function (event) {
      resolve(request.result);
    };
  });
}

async function getCertificateHashesFromCache() {
  const certs = await getValueFromIndexedDB("certificates");
  if (!Array.isArray(certs)) {
    return;
  }
  const nowWithExpirationBuffer = Date.now() + CERTIFICATE_EXPIRATION_BUFFER;
  const isCertActual = certs.some(
    (cert) => nowWithExpirationBuffer < cert.expiresAt
  );
  if (!isCertActual) {
    return;
  }
  return certs;
}

async function setCertificateHashesToCache(certs) {
  return await setValueToIndexedDB("certificates", certs);
}

async function getCertificateHashesFromApi() {
  const resp = await fetch(certificateHashesUrl);
  const respJson = await resp.json();
  const certs = respJson.data;
  if (!Array.isArray(certs)) {
    throw new Error("Failed to get certificates");
  }
  return certs;
}

async function checkAndUpdateCertificates() {
  const cachedCerts = await getCertificateHashesFromCache();
  if (cachedCerts) {
    return cachedCerts;
  }
  const newCerts = await getCertificateHashesFromApi();
  await setCertificateHashesToCache(newCerts);
  return newCerts;
}

async function getCertificateHashes() {
  console.log(`Getting certificate hashes`);
  const certs = await checkAndUpdateCertificates();
  return certs.map(
    ({ hash }) => new Uint8Array(hash.split(",").map((char) => ~~char)).buffer
  );
}

async function createWebTransportClient(url) {
  const hashes = await getCertificateHashes();
  // eslint-disable-next-line no-undef
  const client = new WebTransport(url, {
    serverCertificateHashes: [
      ...hashes.map((value) => ({
        algorithm: "sha-256",
        value,
      })),
    ],
  });
  await client.ready;
  client.closed
    .then(() =>
      console.log(`The HTTP/3 connection to ${url} closed gracefully`)
    )
    .catch((error) =>
      console.error(
        `The HTTP/3 connection to ${url} closed due to error ${error}`
      )
    );
  console.log("CLIENT CREATED");
  return client;
}

function getIsPreferNetwork(event) {
  return (
    event.request.destination === "document" ||
    preferNetworkList.includes(new URL(event.request.url).pathname) ||
    event.request.cache === "no-store" ||
    event.request.cache === "no-cache"
  );
}

async function putInCache(event, response) {
  if (!shouldBeCached(event, response)) return response;

  const clonedResponse = response.clone();
  const cache = await caches.open(WEB_PLAYER_ASSETS_CACHE);
  const key = event.request.url.startsWith("chrome-extension")
    ? "http://fake" + new URL(event.request.url).pathname
    : event.request;
  await cache.put(key, clonedResponse);
}

async function fetchAndCache(event) {
  const response = await fetch(getFetchInput(event.request));
  await putInCache(event, response);
  return response;
}

function getFetchInput(request) {
  // not sure what is happening here
  // we had an issue when an image request was failing with CORS
  // but we had correct s3 config
  // it felt like the browser didn't set origin and cached the response
  return request.url.includes("/processed/")
    ? request.url + `?${Math.random()}`
    : request;
}

async function getFromCache(event) {
  const cached = await caches.match(event.request);
  if (cached) return cached;

  const fakeURL = "http://fake" + new URL(event.request.url).pathname;
  const cached2 = await caches.match(fakeURL);
  if (cached2) return cached2;

  const cached3 = await caches.match(new URL(event.request.url).pathname);
  if (cached3) return cached3;
}

function shouldBeCached(event, response) {
  // eslint-disable-next-line no-restricted-globals
  if (self.location.hostname === "localhost") return false;

  if (event.request.url.startsWith("chrome-extension")) return false;
  if (!response || !response.ok || response.type !== "basic") return false;

  return new URL(event.request.url).pathname.startsWith("/static");
}
